#if !defined(__COEFF_H__)
#define __COEFF_H__

#include <typeinfo>
#include <stdexcept>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <string>
#include <iostream>
//#include <bits/type_traits.h>
#include <memory>
#include <numeric>


#include <functional>
#include <stack>
#include <vector>
#include <ext/numeric>
#include "zp.h"
#include "rational.h"
#include "fpoint.h"
#include "malloc_alloc.h"

#if defined(__SGI__)
#define frexpf frexp
#define ldexpf ldexp
#endif

using std::type_info;

template<class T> struct SafeType {
	static const type_info	&id() {return typeid(T);}
};

template<typename T>
struct SafeType<const T>: public SafeType<T> {};


template<class T> struct ReadableTypeName {
	static const char *name() {return SafeType<T>::id().name();}	
};

template<typename T>
struct ReadableTypeName<const T>: public ReadableTypeName<T> {};

template<> struct ReadableTypeName<float> {
	static const char *name() {return "float";}
};

template<> struct ReadableTypeName<double> {
	static const char *name() {return "double";}
};

template<> struct ReadableTypeName<Rational> {
	static const char *name() {return "Rational";}
};

template<> struct ReadableTypeName<Fpoint> {
	static const char *name();
};

template<> struct ReadableTypeName<Zp> {
	static const char *name();
};


struct Vdomain {

  virtual ~Vdomain() {}
	virtual void destroy()=0;

	virtual Vdomain *clone() const=0;

	virtual void copy(const Vdomain *)=0;

	virtual int sign() const=0;

	virtual bool isZero() const=0;
	virtual bool isUnit() const=0;
//	virtual int degeneracy() const=0;

	virtual Vdomain *zero() const=0;
	virtual Vdomain *unit() const=0;
	virtual Vdomain *random() const=0;

	virtual bool isLess(const Vdomain *) const=0;
	virtual bool isEqual(const Vdomain *) const=0;
	virtual bool exact() const=0;
	virtual int order() const=0;

	virtual Vdomain *add(const Vdomain *) const=0;
	virtual Vdomain *sub(const Vdomain *) const=0;
	virtual Vdomain *mul(const Vdomain *) const=0;
	virtual Vdomain *div(const Vdomain *) const=0;

	virtual void add_auto(const Vdomain *)=0;
	virtual void sub_auto(const Vdomain *)=0;
	virtual void mul_auto(const Vdomain *)=0;
	virtual void div_auto(const Vdomain *)=0;

	virtual int exp2() const=0;
	virtual Vdomain *mul2exp(int e) const=0;
	virtual int edeg(int eb) const=0;

	virtual Vdomain *pow(int) const=0;
	virtual Vdomain *sqrt() const=0;

	virtual void neg_auto()=0;
	virtual void abs_auto()=0;

	virtual Vdomain *neg() const=0;
	virtual Vdomain *abs() const=0;

  virtual void print(std::ostream &os) const=0;
  virtual void scan(std::istream &is)=0;

	virtual double toDouble() const=0;

	virtual const type_info &type() const=0;
	virtual const char *typeName() const=0;

protected:
	Vdomain() = default;
};

template<class T,class POD> struct Opa {

	int sign(const T &data) const {
		if (data<T(0)) return -1;
		if (data==T(0)) return 0;
		return 1;
	}

	bool isZero(const T &data) const {
		return data==T(0);
	}
	bool isUnit(const T &data) const {
		return data==T(1);
	}
	static bool exact() {
		return false;
	}

	T zero(const T &) const {
		return T(0);
	}

	T unit(const T &) const {
		return T(1);
	}

	static T random() {
		return T(0.5+drand48()/2.);
	}

	int order(const T &data) const {
		return 0;
	}

	T pow(const T &data,int k) const {
		if (k>0)
			return __gnu_cxx::power(data,k);
		else
		if (k<0) {
			PRECONDITION(data!=T(0));
			return __gnu_cxx::power(T(1)/data,-k);
		}
		else
			return T(1);
	}
	T sqrt(const T &data) const {
		return ::sqrt(data);
	}

	void neg_auto(T &data) const {
		data=-data;
	}
	void abs_auto(T &data) const {
		if (data<T(0)) neg_auto(data);
	}

	T abs(const T &data) const {
		return  (data<T(0)) ? -data:data;
	}

	void print(std::ostream &os, const T &data) const {
		os<<data;		
	}
	void scan(std::istream &is, T &data) const {
		is>>data;		
	}

	double toDouble(const T &data) const {
		return double(data);
	}
};

template<class T> struct Opa<T,std::false_type> {
	int sign(const T &data) const {
		return data.sign();
	}

	bool isZero(const T &data) const {
		return data.isZero();
	}
	bool isUnit(const T &data) const {
		return data.isUnit();
	}

	T zero(const T &data) const {
		return data.zero();
	}

	T unit(const T &data) const {
		return data.unit();
	}

	static T random() {
		return T::random();
	}

	static bool exact()  {
		return T::exact();
	}
	int order(const T &data) const {
		return 0;
	}

	T pow(const T &data,int k) const {
		return data.pow(k);
	}
	T sqrt(const T &data) const {
		return data.sqrt();
	}

	void neg_auto(T &data) const {
		data.neg_auto();
	}
	void abs_auto(T &data) const {
		data.abs_auto();
	}

	T abs(const T &data) const {
		return data.abs();
	}

	void print(std::ostream &os, const T &data) const {
		data.print(os);
	}
	void scan(std::istream &is, T &data) const {
		data.scan(is);
	}

	double toDouble(const T &data) const {
		return data.toDouble();
	}
};

class MethodIsUndefined {
protected:
	static void undefined(const char *mName, const char *tName) {
		throw std::runtime_error(std::string("The method '")+mName+"' is undefined for the type "+tName);
	}
};

template<class T>
struct TmethodIsUndefined: public MethodIsUndefined {
	static void tUndefined(const char *mName) {
		undefined(mName,ReadableTypeName<T>::name());
	}
};

template<class T> struct BinExp: public TmethodIsUndefined<T> {
  using TmethodIsUndefined<T>::tUndefined;
	static int exp2(const T &) {
		tUndefined("extraction of binary exponent");
		return 0;//to shut up the compiler
	}
	static T mul2exp(const T &,int e) {
		tUndefined("multiplication with a power of 2");
		return T();//to shut up the compiler
	}
};

template<>
struct BinExp<Fpoint> {
	static int exp2(const Fpoint &fp) {
		return fp.exp2();
	}
	static Fpoint mul2exp(const Fpoint &fp,int e) {
		return fp.mul2exp(e);
	}
};

template<>
struct BinExp<float> {
	static int exp2(float f) {
		int e;
		frexpf(f,&e);
		return e;
	}
	static float mul2exp(float f,int e) {
		return ldexpf(f,e);
	}
};

template<>
struct BinExp<double> {
	static int exp2(double d) {
		int e;
		frexp(d,&e);
		return e;
	}
	static double mul2exp(double d,int e) {
		return ldexp(d,e);
	}
};

class EdegBase {
protected:
	static int weightedOrder(int ord,int mant,int w) {
		if (ord>0) return 0;
		ord=-ord;
		return (ord*w)/mant;
	}
};

template<class T>
struct Edeg: public TmethodIsUndefined<T> {
  using TmethodIsUndefined<T>::tUndefined;
	static int edeg(const T &,int eb) {
		tUndefined("e-degree");
    return 0;
	}
};

template<>
struct Edeg<float>: public EdegBase {
	static int edeg(float f,int eb) {
		return weightedOrder(BinExp<float>::exp2(f),23,eb);
	}
};

template<>
struct Edeg<double>: public EdegBase {
	static int edeg(double d,int eb) {
		return weightedOrder(BinExp<double>::exp2(d),52,eb);
    //return weightedOrder(BinExp<double>::exp2(d),45,eb);
		//return std::max(0,-BinExp<double>::exp2(d)/20);
	}
};

template<>
struct Edeg<Fpoint>: public EdegBase {
	static int edeg(const Fpoint &fp,int eb) {
		return weightedOrder(BinExp<Fpoint>::exp2(fp),fp.precision(),eb);
		//return std::max(0,-BinExp<Fpoint>::exp2(fp)/17);
	}
};

template<>
struct Edeg<Rational> {
	static int edeg(const Rational &r,int eb) {
		int ord=r.order();
		return (ord<eb ? ord:(eb-1));
	}
};

template<>
struct Edeg<Zp> {
	static int edeg(Zp,int) {
		return 0;
	}
};

template<class T,class POD = typename std::is_pod<T>::type> struct Tdomain: public Vdomain {

	virtual Vdomain *clone() const {
		return cache().allocate(m_data);
	}

	virtual void copy(const Vdomain *vd) {
		m_data=const_data(vd);
	}

	static Vdomain *create(const T &data) {

		return cache().allocate(data);
	}

	static void assign(Vdomain *vd,const T &d) {
		data(vd)=d;
	}
	virtual void destroy() {
		return cache().deallocate(this);
	}

	virtual const type_info &type() const {return my_type();}
	virtual const char *typeName() const {return my_name();}

	static const type_info &my_type() {return SafeType<T>::id();}
	static const char *my_name() {return ReadableTypeName<T>::name();}

// domain independent methods

	virtual bool isLess(const Vdomain *vd) const {
		if (my_type()!=vd->type()) return false;
		return m_data<const_data(vd);
	}
	virtual bool isEqual(const Vdomain *vd) const {
		if (my_type()!=vd->type()) return false;
		return m_data==const_data(vd);
	}

	virtual Vdomain *add(const Vdomain *vd) const {
		return create(m_data+const_data(vd));		
	}
	virtual Vdomain *sub(const Vdomain *vd) const {
		return create(m_data-const_data(vd));		
	}
	virtual Vdomain *mul(const Vdomain *vd) const {
		return create(m_data*const_data(vd));		
	}
	virtual Vdomain *div(const Vdomain *vd) const {
		return create(m_data/const_data(vd));
	}

	virtual void add_auto(const Vdomain *vd) {
		m_data+=const_data(vd);
	}
	virtual void sub_auto(const Vdomain *vd) {
		m_data-=const_data(vd);
	}
	virtual void mul_auto(const Vdomain *vd) {
		m_data*=const_data(vd);
	}
	virtual void div_auto(const Vdomain *vd) {
		m_data/=const_data(vd);
	}

	virtual Vdomain *neg() const {
		return create(-m_data);
	}

// domain dependent methods

  virtual void print(std::ostream &os) const {
		opa.print(os,m_data);
	}
  virtual void scan(std::istream &is) {
		opa.scan(is,m_data);
	}

	virtual int sign() const {
		return opa.sign(m_data);
	}

	virtual bool isZero() const {
		return opa.isZero(m_data);
	}
	virtual bool isUnit() const {
		return opa.isUnit(m_data);
	}

	virtual Vdomain *zero() const {
		return create(opa.zero(m_data));
	}
	virtual Vdomain *unit() const {
		return create(opa.unit(m_data));
	}

	virtual Vdomain *random() const {
		return create(opa.random());
	}

	virtual bool exact()  const {
		return opa.exact();
	}
	virtual int order() const {
		return opa.order(m_data);
	}

	virtual Vdomain *pow(int k) const {
    return create(opa.pow(m_data,k));
	}
	virtual Vdomain *sqrt() const {
		return create(opa.sqrt(m_data));
	}

	virtual void neg_auto() {
		opa.neg_auto(m_data);
	}
	virtual void abs_auto() {
		opa.abs_auto(m_data);
	}

	virtual Vdomain *abs() const {
		return create( opa.abs(m_data) );
	}

	virtual int exp2() const {
		return binExp.exp2(m_data);
	}

	virtual Vdomain *mul2exp(int e) const {
		return create( binExp.mul2exp(m_data,e) );
	}

	virtual int edeg(int eb) const {
		PRECONDITION(eb>0);
		return eDeg.edeg(m_data,eb);
	}

	virtual double toDouble() const {
		return opa.toDouble(m_data);
	}

	static T &data(Vdomain *vd) {
		aropar(vd);
		return dynamic_cast<Tdomain *>(vd)->m_data;
	}
	static const T &const_data(const Vdomain *vd) {
		aropar(vd);
		return dynamic_cast<const Tdomain *>(vd)->m_data;
	}

private:

	Tdomain(const T &d):m_data(d) {}

	static void aropar(const Vdomain *d) {
		PRECONDITION(d && my_type()==d->type());
	}

	struct Cache;
	friend struct Cache;

	struct Cache {
		Tdomain *allocate() {
			PRECONDITION(valid);
			return fetch();			
		}
		Tdomain *allocate(const T &data) {
			PRECONDITION(valid);
			return fetch(data);			
		}

		void deallocate(Tdomain *p) {
			dispose(p);
		}

		Cache():nd(0),ns(0),nh(0),valid(true),seq(),stack(seq) {seq.reserve(1024);}
		~Cache() {
			int unat=flush();
			valid=false;
			std::cerr<<"Cached "<<typeid(T).name()<<" coeffs stat: dispatched "<<nd<<" stored "<<ns<<" hits "<<nh<<" unattended: "<<unat<<'\n';
		}
	private:
		int nd,ns,nh;
		bool valid;


		typedef std::vector<Tdomain*> Seq;
		typedef std::stack<Tdomain*,Seq> Stack;
		Seq seq;
		Stack stack;

		
		bool empty() const {return stack.empty();}
		bool full() const {return stack.size()==seq.capacity();}
		void push(Tdomain *td) {
			PRECONDITION(!full());
			stack.push(td);						
		}
		Tdomain *pop() {
			PRECONDITION(!empty());
			Tdomain *tmp=stack.top();
			stack.pop();
			return tmp;
		}

		int flush() {
			PRECONDITION(valid);
			int unat=0;
			while (!empty()) {delete pop();++unat;}
			return unat;
		}		


		Tdomain *fetch() {
			++nh;
			if (empty()) return new Tdomain(T());
			++nd;
			return pop();
		}
		Tdomain *fetch(const T &data) {
			++nh;
			if (empty()) return new Tdomain(data);
			++nd;
			Tdomain *tmp = pop();
			tmp->m_data = data;
			return tmp;
		}
		void dispose(Tdomain *p) {
			if (!valid || full())
				delete p;
			else {
				push(p);
				++ns;
			}
		}
	};

	static Cache &cache() {
		static Cache s_cache;
		return s_cache;
	}

	T m_data;
	Opa<T,POD> opa;
	BinExp<T> binExp;
	Edeg<T> eDeg;

  void* operator new(size_t n) { return malloc_alloc::allocate(n);}
  void operator delete(void* p) {malloc_alloc::deallocate(p,sizeof(Tdomain));}
};

class Coeff {

	Vdomain *m_data;

	Vdomain *data() {
		PRECONDITION(m_data);

		return m_data;
	}

	const Vdomain *const_data() const {
		PRECONDITION(m_data);
		return m_data;
	}

	static Vdomain *clone(const Coeff &c) {
		if (c.m_data) return c.m_data->clone();
		return 0;
	}

	Coeff(Vdomain *vd):m_data(vd) {}
	template<class T> Coeff &assign(const T &data) {
		if (m_data) {
			if ( m_data->type()==Tdomain<T>::my_type() ) Tdomain<T>::assign(m_data,data);
			else {
				m_data->destroy();
				m_data = Tdomain<T>::create(data);
			}
		}

		else
			m_data = Tdomain<T>::create(data);
		return *this;
	}

  template<class T> T &raise() {return Tdomain<T>::data(m_data);}
  template<class T> const T &const_raise() const {return Tdomain<T>::const_data(m_data);}

public:

	Coeff():m_data(0) {}
	Coeff(const Coeff &c):m_data( clone(c) ) {}
	Coeff(float f):m_data( Tdomain<float>::create(f) ) {}
	Coeff(double d):m_data( Tdomain<double>::create(d) ) {}
	Coeff(const Rational &r):m_data( Tdomain<Rational>::create(r) ) {}
	Coeff(const Fpoint &fp):m_data( Tdomain<Fpoint>::create(fp) ) {}
	Coeff(Zp z):m_data( Tdomain<Zp>::create(z) ) {}

	~Coeff() {
		if (m_data) m_data->destroy();
	}

	Coeff &operator=(const Coeff &);
	Coeff &operator=(float f) {return assign(f);}
	Coeff &operator=(double d) {return assign(d);}
	Coeff &operator=(const Rational &r) {return assign(r);}
	Coeff &operator=(const Fpoint &fp) {return assign(fp);}
	Coeff &operator=(Zp z) {return assign(z);}

	bool isLess(const Coeff &c) const;
	bool isEqual(const Coeff &c) const;

	bool operator<(const Coeff &c) const {
		return isLess(c);
	}
	bool operator==(const Coeff &c) const {
		return isEqual(c);
	}

	Coeff operator+(const Coeff &c) const {
		return const_data()->add(c.const_data());
	}
	Coeff operator-(const Coeff &c) const {

		return const_data()->sub(c.const_data());
	}
	Coeff operator*(const Coeff &c) const {
		return const_data()->mul(c.const_data());
	}
	Coeff operator/(const Coeff &c) const {
		return const_data()->div(c.const_data());
	}

	Coeff &operator+=(const Coeff &c) {
		data()->add_auto(c.const_data());
		return *this;
	}
	Coeff &operator-=(const Coeff &c) {
		data()->sub_auto(c.const_data());
		return *this;
	}
	Coeff &operator*=(const Coeff &c) {
		data()->mul_auto(c.const_data());
		return *this;
	}
	Coeff &operator/=(const Coeff &c) {
		data()->div_auto(c.const_data());
		return *this;
	}

	int exp2() const {
		return const_data()->exp2();
	}
	Coeff mul2exp(int e) const {
		return const_data()->mul2exp(e);
	}

	Coeff zero() const {
		return Coeff(const_data()->zero());
	}
	Coeff unit() const {
		return Coeff(const_data()->unit());
	}
	Coeff random() const {
		return Coeff(const_data()->random());
	}

	int sign() const {
		return const_data()->sign();
	}

	bool isZero() const {
		return const_data()->isZero();
	}
	bool isUnit() const {
		return const_data()->isUnit();
	}
	bool exact()  const {
		return const_data()->exact();
	}
	int order() const {
		return const_data()->order();
	}

	template<class T> bool isType() const {return m_data && m_data->type()==Tdomain<T>::my_type();}

	Coeff pow(int k) const {
		return Coeff(const_data()->pow(k));
	}
	Coeff sqrt() const {
		return Coeff(const_data()->sqrt());
	}

	void neg_auto() {
		data()->neg_auto();
	}
	void abs_auto() {
		data()->abs_auto();
	}

	Coeff operator-() const {
		return Coeff(const_data()->neg());
	}
	Coeff abs() const {
		return Coeff(const_data()->abs());
	}

	double toDouble() const {
		return const_data()->toDouble();
	}

	int edeg(int eb) const {
		return const_data()->edeg(eb);
	}

  void print(std::ostream &os) const;
  void scan(std::istream &is);

	const type_info &type() const;
	const char *typeName() const;

  operator float &() {return raise<float>();}
  operator float() const {return const_raise<float>();}

  operator double &() {return raise<double>();}
  operator double() const {return const_raise<double>();}

  operator Rational &() {return raise<Rational>();}
  operator const Rational &() const {return const_raise<Rational>();}

  operator Fpoint &() {return raise<Fpoint>();}
  operator const Fpoint &() const {return const_raise<Fpoint>();}

  operator Zp &() {return raise<Zp>();}
  operator Zp() const {return const_raise<Zp>();}

	static void randomize() {
		srand48(time(0));
		std::srand(time(0));
	}
};


inline std::ostream &operator<<(std::ostream &os,const Coeff &c) {
	c.print(os);
	return os;
}

inline std::istream &operator>>(std::istream &is,Coeff &c) {
	c.scan(is);
	return is;
}

inline bool operator!=(const Coeff &x, const Coeff &y) {
  return !(x ==y);
}

inline bool operator>(const Coeff &x, const Coeff &y) {
  return y<x;
}

inline bool operator<=(const Coeff &x, const Coeff &y) {
  return !(y < x);
}

inline bool operator>=(const Coeff &x, const Coeff &y) {
  return !(x < y);
}

struct AlmostEqual {

	enum Scheme {Exact=0,Abs=1,Rel=2};

	AlmostEqual(const Coeff &eps):m_epsilon(eps),m_scheme(guessScheme(eps)) {}
	AlmostEqual(const Coeff &eps,Scheme scm):m_epsilon(eps),m_scheme(scm) {}
	AlmostEqual(const AlmostEqual &aeq):m_epsilon(aeq.m_epsilon),m_scheme(aeq.m_scheme) {}

	AlmostEqual &operator=(const AlmostEqual &aeq) {
		m_epsilon=aeq.epsilon();
		m_scheme=aeq.scheme();
		return *this;
	}

	const Coeff &epsilon() const {return m_epsilon;}
	void epsilon(const Coeff &eps) {m_epsilon=eps;}

	Scheme scheme() const {return m_scheme;}
	void scheme(const Scheme scm) {m_scheme=scm;}

	bool isEqual(const Coeff &c1, const Coeff &c2) const;
	bool isZero(const Coeff &c) const;
		
	bool operator()(const Coeff &c1, const Coeff &c2) const {
    if (&c1==&c2) return true;
		return isEqual(c1,c2);
	}
	bool operator()(const Coeff &c) const {
		return isZero(c);
	}

	void print (std::ostream &os) const;

	static Scheme guessScheme(const Coeff &c) {return c.exact()?Exact:Rel;}
	static const char *schemeName(Scheme scm);

private:

	Coeff m_epsilon;
	Scheme m_scheme;
	
};

inline std::ostream &operator<<(std::ostream &os,const AlmostEqual &ae) {
	ae.print(os);
	return os;
}

#endif // __COEFF_H__
