#if !defined(__CSET_H__)
#define __CSET_H__

#include <ostream>
#include <set>
#include <map>
#include <vector>
#include <functional>
#include <algorithm>
#include "checks.h"
#include "basis.h"
#include "report.h"

template<class K,class D>
class Cbase: public std::map<K,D> {
public:
  typedef typename std::map<K,D> Inherited;
  using Inherited::end;
  using Inherited::begin;
	typedef typename Inherited::const_iterator const_iterator;
private:
	struct PQgreater: public std::binary_function<const_iterator,const_iterator,bool> {
		bool operator()(const_iterator i1,const_iterator i2) {
			return i2->second<i1->second || (i2->second==i1->second && &*i2<&*i1);
		}
	};
	struct PQ: public std::set<const_iterator,PQgreater> {
		typedef typename Cbase::const_iterator cbase_iterator;
		typedef std::set<cbase_iterator,PQgreater> Inherited;
		typedef typename Inherited::const_iterator const_iterator;
		typedef typename Inherited::const_reverse_iterator const_reverse_iterator;
		typedef const_iterator pq_iterator;
		typedef const_reverse_iterator pq_reverse_iterator;
	};
	

	PQ m_pq;
public:

	typedef typename PQ::pq_iterator pq_iterator;
	typedef typename PQ::pq_reverse_iterator pq_reverse_iterator;

  typedef typename Inherited::value_type value_type;

	void insert(const value_type &v) {
		std::pair<const_iterator,bool> result=Inherited::insert(v);
		PRECONDITION(result.second);
		m_pq.insert(result.first);
	}

	void insert(const K &k, const D &d) {
		std::pair<const_iterator,bool> result=Inherited::insert(value_type(k,d));
		PRECONDITION(result.second);
		m_pq.insert(result.first);
	}
	void erase(const K &k) {
		const_iterator pos=find(k);
		if (pos!=end()) {
			m_pq.erase(pos);
			Inherited::erase(pos);
		}
	}
	void erase(const_iterator pos) {
		int result = m_pq.erase(pos);
		result+=Inherited::erase(pos->first);
		PRECONDITION(result==2);
	}

	static bool tie(const value_type &v1,const value_type &v2) {
		return v1.second.equal1(v2.second);
	}

	pq_iterator pq_begin() const {return m_pq.begin();}
	pq_iterator pq_end() const {return m_pq.end();}

	pq_reverse_iterator pq_rbegin() const {return m_pq.rbegin();}
	pq_reverse_iterator pq_rend() const {return m_pq.rend();}
};

template<class T>
class Npair: public std::pair<T,T> {
public:
  typedef std::pair<T,T> Inherited;
  using Inherited::first;
  using Inherited::second;
	Npair(T i,T j):Inherited(i,j) {
		PRECONDITION(i!=j);
		if (j<i) std::swap(first,second);
	}
};


class HM1: std::pair<Term,Coeff> {
	int m_ediff;
public:
	typedef std::pair<Term,Coeff> Inherited;
  using Inherited::first;
  using Inherited::second;
	HM1(const Tpolin &bp):
				Inherited( bp.lt().main(),bp.lc().abs() ),
				m_ediff(bp.context().eb()-bp.lt().edegree()) {}
	
	const Term &tlt() const {return first;}
	int ediff() const {return m_ediff;}
	const Coeff &abslc() const {return second;}

	bool less1(const HM1 &hm) const {
		return ediff()<hm.ediff() ||
							(ediff()==hm.ediff() && tlt().degree()<hm.tlt().degree());
	}
	bool equal1(const HM1 &hm) const {return ediff()==hm.ediff() && tlt().degree()==hm.tlt().degree();}

	bool less2(const HM1 &hm) const {return hm.abslc()<abslc();}
	bool equal2(const HM1 &hm) const {return hm.abslc()==abslc();}

	bool less(const HM1 &hm) const {return less1(hm) || (equal1(hm) && less2(hm));}
	bool equal(const HM1 &hm) const {return equal1(hm) && equal2(hm);}

	bool operator<(const HM1 &hm) const {
		return less(hm);
	}
	bool operator==(const HM1 &hm) const {
		return equal(hm);
	}
};

struct HM2: public std::pair<Eterm,Coeff> {
	typedef std::pair<Eterm,Coeff> Inherited;
	HM2(const Tpolin &bp1, const Tpolin &bp2):
			Inherited( bp1.lt().lcm(bp2.lt()), std::max(bp1.lc().abs(),bp2.lc().abs()) ) {}

	const Eterm &lcm() const {return first;}
	const Coeff &abslc() const {return second;}

	bool less1(const HM2 &hm) const {return lcm().edegree()<hm.lcm().edegree() || (lcm().edegree()==hm.lcm().edegree() && lcm().tdegree()<hm.lcm().tdegree()) ;}
	bool equal1(const HM2 &hm) const {return lcm().edegree()==hm.lcm().edegree() && lcm().tdegree()==hm.lcm().tdegree();}

	bool less2(const HM2 &hm) const {return hm.abslc()<abslc();}
	bool equal2(const HM2 &hm) const {return hm.abslc()==abslc();}

	bool less(const HM2 &hm) const {return less1(hm) || (equal1(hm) && less2(hm));}
	bool equal(const HM2 &hm) const {return equal1(hm) && equal2(hm);}

	bool operator<(const HM2 &hm) const {
		return less(hm);
	}
	bool operator==(const HM2 &hm) const {
		return equal(hm);
	}
};

struct Eorbit: public Cbase<int,HM1> {
	typedef Cbase<int,HM1> Inherited;
	typedef Inherited::mapped_type mapped_type;
	typedef mapped_type data_type;

	void insert(const value_type &v) {
		Inherited::insert(v);
	}

	void insert(int k, const Tpolin &bp) {
		Inherited::insert(k,data_type(bp));
	}
};

struct Spairs: public Cbase<Npair<int>,HM2> {
	typedef Cbase<Npair<int>,HM2> Inherited;
	typedef Inherited::mapped_type mapped_type;
	typedef mapped_type data_type;

	void insert(const value_type &v) {
		Inherited::insert(v);
	}

	void insert(int i,int j,const Tpolin &bp1, const Tpolin &bp2) {
		Inherited::insert(key_type(i,j),data_type(bp1,bp2));
	}
};

class Cset: public Report, public Stat {
	std::vector<Tpolin> m_table;
	Eorbit m_eorbit;
	Spairs m_spairs;
	Eorbit m_edone;
	Spairs m_sdone;
	mutable Pcounter m_ecounter;
	mutable Pcounter m_scounter;

public:
	typedef Eorbit::value_type Eshift;
	typedef Spairs::value_type Spair;

	typedef Eorbit::key_type eshift_key;
	typedef Spairs::key_type spair_key;

	typedef Eorbit::const_iterator eorbit_iterator;
	typedef Spairs::const_iterator spairs_iterator;

private:

	void print(std::ostream &os,const Eshift &es) const;
	void print(std::ostream &os,const Spair &sp) const;
	void print(std::ostream &os,const Eorbit &es) const;
	void print(std::ostream &os,const Spairs &s) const;

	void printBrief(std::ostream &os,const Eshift &es) const;
	void printBrief(std::ostream &os,const Spair &sp) const;

	void printBrief(std::ostream &os,const Eorbit &eorb) const;
	void printBrief(std::ostream &os,const Spairs &s) const;


public:
	Cset(const Report &r, const Stat &s):Report(r), Stat(s), m_ecounter('E'), m_scounter('S') {
		stat().add("sp","S-pairs delivered for reduction");
		stat().add("es","E-shifts delivered for reduction");
		stat().add("bb1es","E-shifts excluded by coprime");
		stat().add("bb1sp","S-pairs excluded by coprime");
		stat().add("bb2es","E-shifts excluded by chain");
		stat().add("bb2sp","S-pairs excluded by chain");
	}
	Tpolin eval(const Eshift &eshift) const;
	std::pair<Tpolin,Tpolin> eval(const Spair &spair) const;

  void add(const Basis &b);

	bool empty() const {return m_spairs.empty() && m_eorbit.empty();}
	void clear();

	int deliver(Basis &b);

	void print(std::ostream &os) const;
	void printBrief(std::ostream &os) const;

private:
  std::pair<int,int> add(const Tpolin &bp);

	int bb2eorb(int shed);
	int bb2sp(int shed);

	int bb2(int shed);
};


inline std::ostream &operator<<(std::ostream &os, const Cset &cs) {
	cs.print(os);
	return os;
}

#endif //__CSET_H__
