#include <fstream>
#include <list>
#include "inputparser.h"
#include "gbasis.h"
#include "borderbasis.h"
#include "multmatr.h"
#include "printSequence.h"

using namespace std;


struct ValTerm: public Term {
	const Term &term() const {return *this;}
	ValTerm(const Term &t):Term(t) {}
	std::ostream &print(std::ostream &os) const {
		return os<<"val_"<<term();
	}
};

inline std::ostream &operator<<(std::ostream &os, const ValTerm &vt) {
	return vt.print(os);
}

struct SubsTerm: public ValTerm {
	const ValTerm &valTerm() const {return *this;}
	SubsTerm(const ValTerm &vt):ValTerm(vt) {}
	std::ostream &print(std::ostream &os) const {
		return os<<term()<<"="<<valTerm();
	}
};

typedef std::set<SubsTerm, std::greater<Term> > SubsPend;

inline std::ostream &operator<<(std::ostream &os, const SubsTerm &vt) {
	return vt.print(os);
}

int main(int argc, char *argv[]) {
  try {
    Zp::setmod(65521);
    const char *ifile = (argc>1 ? argv[1]:"ideal.txt");
    ifstream is(ifile);
    if (!is) {
      cout<<"Can not open input file "<<ifile<<'\n';
      return 1;
    }
    InputParser ip(is);
    if ( ip() ) {
      const InputParser::Results &input = ip.results();
      Report report(input["report"].cast<Report>());
      const InputParser::Polins &pols= input["ideal"].cast<InputParser::Polins>();
      report.rs().flush();
      Pcounter bcount('F');
      Basis b;
      for(InputParser::Polins::const_iterator i=pols.begin();i!=pols.end();++i) {				
        Polin p=*i;
        p.normalize();
        p.calibrateEdegree();
        if (!p.empty()) b.push_back(Tpolin(p,bcount));
      }
    report.rs()<<input["context"]<<"The normalized and translated input basis contains "<<b.size()<<" polynomials:\n"<<b;
    Gbasis gb(report);
    gb(b);
    cout<<"\nGB complete.\n";
    report.rs()<<"The reduced Groebner basis contains "<<b.size()<<" polynomials:\n"<<b;
    bool rem=false;
    for (Basis::const_iterator i=b.begin();i!=b.end();) {
      int eord=i->eorder();
      if (eord>0) {
        report.rs()<<"Removing e-order="<<eord<<" element "<<i->tag()<<" from the basis\n";
        b.erase(i++);
        rem=true;
      }
#if 1			
      else
      if (i->isAlmostZero()) {
        report.rs()<<"Removing the element "<<i->tag()<<" with tiny coefficients\n";
        b.erase(i++);
        rem=true;
      }
#endif			
      else ++i;
    }
    if (rem) {
      report.rs()<<"The Groebner basis with positive e-order and tiny coeffs elements removed contains "<<b.size()<<" polynomials:\n"<<b;
    }
    if (b.isStable()) {
      report.rs()<<"The Groebner basis is stable\n";
      if (b.isZeroDim()) {
        report.rs()<<"The ideal is zero-dimensional\n";
        SetOfTerms ns=b.nset();
        report.rs()<<"The normal set is "<<ns<<'\n';
        report.rs()<<"The system has "<<ns.size()<<" solutions\n";
        cout<<ns.size()<<" solutions\n";
        SetOfTerms bs=ns.bset();
        report.rs()<<"The border set contains "<<bs.size()<<" terms: "<<bs<<'\n';
        BorderBasis bb;
        SetOfTerms degen;
        Pcounter bcount('B');
        for (SetOfTerms::const_iterator t=bs.begin();t!=bs.end();++t) {
          Polin f(*t);
          for(size_t i=0;i<f.size();) {
            const Eterm s=f.monom(i).eterm();
            Basis::div_iterator first=b.lower_divisor(s);
            Basis::div_iterator last=b.upper_divisor(s);
            if (first!=last) {
              Basis::const_iterator q=first;
              PRECONDITION(q!=b.end());
              while (++first!=last) {
                if (q->lc().abs()<first->lc().abs()) q=first;
              }
              const Coeff c=-f.monom(i).coeff()/q->lc();
              f=f.eliminate(*q*(s.divide(q->lt())),c,s);		
            }
            else {
              ++i;
            }
          }
          Basis::const_iterator g=find_if( b.begin(),b.end(), __gnu_cxx::compose1(bind2nd(equal_to<Term>(),*t),__gnu_cxx::compose1(mem_fun_ref(&Eterm::main),mem_fun_ref(&Polin::lt))));
          f=Polin(*t)-f;
          //report.rs()<<"IB: "<<f<<'\n';
          f.resetEdegree();
          Tpolin be = (g==b.end() ? Tpolin(f,bcount):Tpolin(f,g->tag()));
          if (be.monom(*t)==be.const_end()) {
            degen.insert(*t);
          }
          else {
            bb.add(*t,be);
          }
        }
        bb.check(ns,bs,report);
        cout<<"Border basis complete.\n";
        report.rs()<<"The border basis with "<<bb.size()<<" polynomials is "<<(degen.empty()? "":"degenerated")<<'\n'<<bb;
        if (!degen.empty()) {
          report.rs()<<"The following border terms vanish "<<degen;
        }                            
        else
        if (input["mm"].cast<string>()=="yes")	{
          MultMatr em(ns);
          const PolyContext &ctx=input["context"];
          size_t n=ctx.abc().size();
          list<Coeff> rc;
          list<string> mnames;
          Coeff::randomize();
          for (size_t i=0;i<n;++i) {
            MultMatr mm(ns);
            mm(bb,i);    
            report.rs()<<"The multiplication with '"<<ctx.abc()[i]<<"' matrix M_"<<ctx.abc()[i]<<" is\n"<<mm;
            report.rs()<<"The maple form is\n";mm.printMaple(report.rs());
            rc.push_back(ctx.coeff().random());
            mnames.push_back(string("M_")+ctx.abc()[i]);
            em.comb(mm,rc.back());
          }
          report.rs()<<"\nThe matrix for eigenvector computation is ";
          printRow(report.rs(),rc.begin(),rc.end());
          report.rs()<<" * ";
          printCol(report.rs(),mnames.begin(),mnames.end());
          report.rs()<<" =\n";
          report.rs()<<em;
          report.rs()<<"The maple form is\n";em.printMaple(report.rs());
          cout<<"Multiplication matrices complete.\n";
          if (input["maple"].cast<string>()=="yes") {
            string maplefName=report.fileName();
						size_t dotPos=maplefName.rfind('.');
						if (dotPos==std::string::npos) 
							maplefName+=".maple";			
						else							
            	maplefName.replace(dotPos,maplefName.size()-dotPos,".maple");
            ofstream ms(maplefName.c_str());
						ms.precision(report.rs().precision());
						if (input.find("digits")!=input.end()) ms<<"Digits:="<<int(input["digits"])<<":\n";
            ms<<"with(linalg):\n";
            ms<<"F:=";printCol(ms,pols.begin(),pols.end());ms<<":\n";
            ms<<"M:=";em.printMaple(ms);ms<<":\n";
            ms<<"E:=eigenvectors(M):\nZ:=[]:\n";
            ms<<"for i in E do\n";
						SubsPend subsList;
            for (size_t i=0;i<n;++i) {
							Term tvar=Term(ctx).var(i);
              SetOfTerms::const_iterator varPos=ns.find(tvar);
              if (varPos!=ns.end()) {
								ValTerm valTerm=*varPos;
              	ms<<"  "<<valTerm<<":=";
								ms<<"op(i[3])["<<(distance(ns.begin(),varPos)+1)<<"]/op(i[3])[1]:\n";
								subsList.insert(valTerm);
							}
            }						
            for (size_t i=0;i<n;++i) {
							Term tvar=Term(ctx).var(i);
              SetOfTerms::const_iterator varPos=ns.find(tvar);
              if (varPos==ns.end()) {
								BorderBasis::const_iterator ltPos=bb.find(tvar);
								PRECONDITION(ltPos!=bb.end());
								ValTerm valTerm=ltPos->first;
								ms<<"  "<<valTerm<<":=";
								if (!subsList.empty()) {
									PrintSequence<SubsPend::const_iterator>  ps(",","subs(",",");
									ps(ms,subsList.begin(),subsList.end());
									ms<<(Polin(ltPos->first)-ltPos->second)<<"):\n";
									subsList.insert(valTerm);
								}
							}
            }
            ms<<"  Z:=[op(Z),";
						PrintSequence<SubsPend::const_iterator> ps(",","[","]");
						ps(ms,subsList.begin(),subsList.end());
            ms<<"]:\n";
            ms<<"od:\n";
            ms<<"for sol in Z do subs(sol,F) od;\n";
          }
        }
      }
      else {
        report.rs()<<"The ideal is not zero-dimensional\n";
      }
    }
    else {
      report.rs()<<"The Groebner basis is not stable\n";
    }
    cout<<"Bye.\n";
    }
    else {
      int l=ip.error().line();
      string errMsg =ip.error().get();
      cout<<"Parse error: "<<errMsg<<" in line "<<l<<'\n';
      string errLine=ip.inputLine(l);
      if (!errLine.empty()) {
        cout<<errLine<<'\n';
        int c=ip.error().col();			
        for (int i=0;i<c;++i) cout<<' ';
        cout<<"^\n";
      }
    }
  }
  catch (runtime_error &re) {
    cerr<<"The fallen victim of abortion: "<<re.what()<<'\n';
  }
  return 0;
}
