#if !defined __RATIONAL_H__
#define __RATIONAL_H__

#include <iostream>
#include <cstdlib>
#include "integer.h"
#include "lrand48.h"


class Rational;

class Rational_error : public std::runtime_error {
  static const char * const prefix;
public:
  explicit Rational_error (const std::string& what_arg): std::runtime_error (std::string(prefix)+what_arg) { }
};

class Rational {

  mpq_t rep;

public:

  Rational() {
		mpq_init(rep);
  }

  Rational(const Rational& r) {
		mpq_init(rep);
		mpq_set(rep, r.rep);
  }

  Rational(const Integer& num) {
		mpz_init_set(mpq_numref(rep), num.rep);
		mpz_init_set_ui(mpq_denref(rep), 1);
  }

  Rational(long num) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_ui(mpq_denref(rep), 1);
  }

  Rational(int num) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_ui(mpq_denref(rep), 1);
  }

  Rational(const Integer& num, const Integer& den) {   
		mpz_init_set(mpq_numref(rep), num.rep);
		mpz_init_set(mpq_denref(rep), den.rep);
		mpq_canonicalize(rep);
  }


  Rational(long num, long den) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(int num, int den) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(const Integer& num, long den) {
		mpz_init_set(mpq_numref(rep), num.rep);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(long num, const Integer& den) {   
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set(mpq_denref(rep), den.rep);
		mpq_canonicalize(rep);
  }

  Rational(const Integer& num, int den) {
		mpz_init_set(mpq_numref(rep), num.rep);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(int num, const Integer& den) {   
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set(mpq_denref(rep), den.rep);
		mpq_canonicalize(rep);
  }

  Rational(long num, int den) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(int num, long den) {
		mpz_init_set_si(mpq_numref(rep), num);
		mpz_init_set_si(mpq_denref(rep), den);
		mpq_canonicalize(rep);
  }

  Rational(const double d) {
		mpq_init(rep);
		*this=d;
  }
  
  explicit Rational(const char* s) {
    mpq_init(rep);
    try {
      set(s);
    }
    catch (const Rational_error &) {
      mpq_clear(rep);
      throw;
    }
  }

  friend Integer& Integer::operator=(const Rational& r);

  ~Rational() {
		mpq_clear (rep);
  }

  Integer num() const {
		return mpq_numref(rep);
  }

  Integer denom() const {
		return mpq_denref(rep);
  }

  void set(const Integer& num, const Integer& den) {
		mpq_set_num(rep, num.rep);
		mpq_set_den(rep, den.rep);
		mpq_canonicalize(rep);
  }
  
  void set_num(const Integer& num) {
		mpq_set_num(rep, num.rep);
		mpq_canonicalize(rep);
  }

  void set_denom(const Integer& den) {
		mpq_set_den(rep, den.rep);
		mpq_canonicalize(rep);
  }


  Rational& set(const char *s) noexcept(false);

  Rational& operator=(const Rational& a) {
		mpq_set(rep, a.rep);
		return *this;
  }

  Rational& operator=(const Integer& a) {
		mpz_set(mpq_numref(rep), a.rep);
		mpz_set_ui(mpq_denref(rep), 1);
		return *this;
  }

  Rational& operator=(long i) {
		mpq_set_si(rep, i, 1);
		return *this;
  }

  Rational& operator= (int i) {
		mpq_set_si(rep, i, 1);
		return *this;
  }

  Rational& operator=(double d);
 
  int sign() const {
    return mpq_sgn(rep);
  }

  bool operator!() const {
		return !mpq_sgn(rep);
  }

  explicit operator void*() const {
		return reinterpret_cast<void*>(mpq_sgn(rep));
  }

  std::ostream &print(std::ostream &os) const;
  std::istream &scan(std::istream &is);

  Rational operator+(const Rational &q) const {
		Rational r;
  	mpq_add(r.rep, rep, q.rep);
		return r;
	}

  Rational &operator+=(const Rational &q) {
    mpq_add(rep, rep, q.rep);
    return *this;  
  }

  Rational operator-(const Rational &q) const {
		Rational r;
  	mpq_sub(r.rep, rep, q.rep);
		return r;
	}

  Rational &operator-=(const Rational &z) {
    mpq_sub(rep, rep, z.rep);
    return *this;  
  }

  Rational operator-() const {
		Rational r;
  	mpq_neg(r.rep, rep);
		return r;
	}
	
	Rational &neg_auto() {
  	mpq_neg(rep, rep);
		return *this;
	}

	Rational abs() const {
		Rational r;
   	mpz_abs(mpq_numref(r.rep), mpq_numref(rep));
   	mpz_abs(mpq_denref(r.rep), mpq_denref(rep));
		return r;
	}

	Rational &abs_auto() {
   	mpz_abs(mpq_numref(rep), mpq_numref(rep));
   	mpz_abs(mpq_denref(rep), mpq_denref(rep));
		return *this;
	}

  Rational operator*(const Rational &q) const {
		Rational r;
  	mpq_mul(r.rep, rep, q.rep);
		return r;
	}

  Rational &operator*=(const Rational &q) {
    mpq_mul(rep, rep, q.rep);
    return *this;
  }

  Rational operator/(const Rational &q) const {
		Rational r;
  	mpq_div(r.rep, rep, q.rep);
		return r;
	}

  Rational &operator/=(const Rational &q) {
		mpq_div(rep, rep, q.rep);
		return *this;
  }
  
  static Rational zero() {return __gnu_cxx::identity_element(std::plus<Rational>());}
  static Rational unit() {return __gnu_cxx::identity_element(std::multiplies<Rational>());}
	static Rational random() {return Rational(1,2)+Rational(int(lrand48()%5000)),int(lrand48()%10000);}

  bool isZero() const {return *this==zero();}
  bool isUnit() const {return *this==unit();}

  int compare(const Rational &q) const { return mpq_cmp(rep, q.rep); }

  bool isEqual(const Rational &q) const {return compare(q)==0;}
  bool isLess(const Rational &q) const  {return compare(q)<0;}

  bool operator==(const Rational &q) const {
		return isEqual(q);
  }

  bool operator<(const Rational &q) const {
		return isLess(q);
  }

  static bool exact() {return true;}

  int order() const {
		int n=mpz_size(mpq_numref(rep))/4;
		int d=mpz_size(mpq_denref(rep))/4;
		return (n<d) ? d:n;
  }

	Integer num_gcd(const Rational &q) const {
  	return num().gcd(q.num());
	}

	Integer denom_gcd(const Rational &q) const {
		return denom().gcd(q.denom());
	}

	Rational nd_gcd(const Rational &q) const {
  	return num_gcd(q),denom_gcd(q);
	}

	Integer num_lcm(const Rational &q) const {
  	return num().lcm(q.num());
	}

	Integer denom_lcm(const Rational &q) const {

  	return denom().lcm(q.denom());
	}

	Rational nd_lcm(const Rational &q) const {
  	return Rational(num_lcm(q),denom_lcm(q));
	}


	Rational n_gcd_d_lcm(const Rational &q) const {
  	return Rational(num_gcd(q),denom_lcm(q));
	}

	Rational pow(long k) const;


	Rational sqrt() const {
  	return Rational(num().sqrt(),denom().sqrt());
	}

  double toDouble() const {
		return mpq_get_d(rep);
  }
};


inline std::ostream &operator<<(std::ostream &os,const Rational &q) {return q.print(os);}
inline std::istream &operator>>(std::istream &is,Rational &q)     {return q.scan(is);}


inline Integer& Integer::operator=(const Rational &q) {
  mpz_tdiv_q(rep, mpq_numref(q.rep), mpq_denref(q.rep));
  return *this;
}

inline Integer::Integer(const Rational& q) {
  mpz_init(rep);
  *this=q;
}

inline bool operator!=(const Rational &x, const Rational &y) {
  return !(x ==y);
}

inline bool operator>(const Rational &x, const Rational &y) {
  return y<x;
}

inline bool operator<=(const Rational &x, const Rational &y) {
  return !(y < x);
}

inline bool operator>=(const Rational &x, const Rational &y) {
  return !(x < y);
}

#endif //__RATIONAL_H__

