#if !defined __ZP_H__
#define __ZP_H__

#include <stdexcept>
#include <cmath>
#include <cstdlib>
#include <functional>
#include <iostream>
#include "checks.h"
#include "lrand48.h"

class Zp_error : public std::runtime_error {
  static const char * const prefix;
public:
  explicit Zp_error (const std::string& what_arg): std::runtime_error (std::string(prefix)+what_arg) { }
};

class Zp {

  static int p;
  static int *inv;

  int k;
  
  static inline int  normalize(int i,int m) {
    PRECONDITION(m);
    if (i>=m || i<-m) i%=m;
    if (i<0) i+=m;
    return i;
  }

  static inline int  normalize(int i) {
    return normalize(i,p);
  }

  static inline int abs(int i) {return i<0?-i:i;}
  
  static inline int mul_int(int i,int j) {
    if (abs(i)<32768 && abs(j)<32768) return i*j;
    return ((long long)i*j)%p;
  }

  enum Internal {internal};

  Zp(int i,Internal):k(i) {}

public:

  static bool setmod(int m);
  static int mod() {return p;}

  Zp():k(0) {}
  Zp(const Zp &z) = default;
  explicit Zp(const char *s):k(-1) {
		set(s);
  }

	Zp(int i):k(normalize(i)) {}

  Zp &operator=(const Zp &z) = default;

  template<class T> Zp &operator=(const T &t) {
		k=normalize(t);
		return *this;
  }

  Zp &operator=(const char *s) {
		set(s);
		return*this;
  }

  Zp &set(const char *s) noexcept(false);

  int sign() const {return k ? (k<=p/2 ? 1:-1):0;}

  int toInt() const {return (sign()<0 ? k-p:k);}
  double toDouble() const {return double(toInt());}

  bool operator!() const {
		return k==0;
  }

  explicit operator void*() const {
		return (void*)(k!=0);
  }
  
  std::ostream &print(std::ostream &os) const;
  std::istream &scan(std::istream &is) noexcept(false);

  Zp operator+(const Zp &z) const {return k+z.k;}
  Zp &operator+=(const Zp &z) {k=normalize(k+z.k);return *this;}
  

  Zp operator-(const Zp &z) const {return k-z.k;}
  Zp &operator-=(const Zp &z) {k=normalize(k-z.k);return *this;}

  Zp operator-() const {return Zp(p-k,internal);}
  Zp &neg_auto() {k=p-k;return *this;}

  Zp abs() const {return sign()<0 ? p-k:k;}
  Zp &abs_auto() {return *this=abs();}


  Zp operator*(const Zp &z) const {return mul_int(k,z.k);}
  Zp &operator*=(const Zp &z) {k=normalize(mul_int(k,z.k));return *this;}

  Zp operator/(const Zp &z) const {return mul_int(k,inv[z.k]);}
  Zp &operator/=(const Zp &z) {k=normalize(mul_int(k,inv[z.k]));return *this;}
  
  Zp &inc() {k=normalize(++k);return *this;}
  Zp &dec() {k=normalize(--k);return *this;}

  Zp &operator++() {return inc();}
  Zp operator++(int) {
		Zp tmp=*this;
		inc();
		return tmp;
  }

  Zp &operator--() {return dec();}
  Zp operator--(int) {
		Zp tmp=*this;
		dec();
		return tmp;
  }


  Zp pow(long e) const;
	Zp sqrt() const {return Zp(int(::sqrt(k)),internal);}
  
  static Zp zero() {return Zp(0,internal);}
  static Zp unit() {return Zp(1,internal);}
  static Zp random() {return Zp(1+lrand48()%(p-1),internal);}

  bool isZero() const {return k==0;}
  bool isUnit() const {return k==1;}

  bool isEqual(const Zp &z) const {return k==z.k;}
  bool isLess(const Zp &z) const {return k<z.k;/*(*this-z).sign()<0;*/}

  bool operator==(const Zp &z) const {
		return isEqual(z);
  }

  bool operator<(const Zp &z) const {
		return isLess(z);
  }

  static bool exact() {return true;}

  static void printInv(std::ostream &os);
};

inline bool operator!=(const Zp &x, const Zp &y) {
  return !(x ==y);
}

inline bool operator>(const Zp &x, const Zp &y) {
  return y<x;
}

inline bool operator<=(const Zp &x, const Zp &y) {
  return !(y < x);
}

inline bool operator>=(const Zp &x, const Zp &y) {
  return !(x < y);
}

template<>
inline Zp __gnu_cxx::identity_element(std::plus<Zp>) {
  return Zp::zero();
}

template<>
inline Zp __gnu_cxx::identity_element(std::multiplies<Zp>) {
  return Zp::unit();
}

inline std::ostream &operator<<(std::ostream &os,const Zp &z) {return z.print(os);}
inline std::istream &operator>>(std::istream &is,Zp &z) {return z.scan(is);}

#endif //__ZP_H__
